<div align="center">

# SSPINNcode

<a href="https://pytorch.org/get-started/locally/"><img alt="PyTorch" src="https://img.shields.io/badge/PyTorch-ee4c2c?logo=pytorch&logoColor=white"></a>
<a href="https://pytorchlightning.ai/"><img alt="Lightning" src="https://img.shields.io/badge/-Lightning-792ee5?logo=pytorchlightning&logoColor=white"></a>
<a href="https://hydra.cc/"><img alt="Config: Hydra" src="https://img.shields.io/badge/Config-Hydra-89b8cd"></a>
<a href="https://github.com/ashleve/lightning-hydra-template"><img alt="Template" src="https://img.shields.io/badge/-Lightning--Hydra--Template-017F2F?style=flat&logo=github&labelColor=gray"></a><br>
[![Paper](http://img.shields.io/badge/paper-arxiv.1001.2234-B31B1B.svg)](https://www.nature.com/articles/nature14539)
[![Conference](http://img.shields.io/badge/AnyConference-year-4b44ce.svg)](https://papers.nips.cc/paper/2020)

</div>

## Description

Code for the ICLR submission: "A Self-Supervised PINN for Inertial Pose and Dynamics Estimation". 
We included video for walking and running is included.

## Credits
This repo is based on the template shared by: https://github.com/ashleve/lightning-hydra-template
We thank the authors for making their template publicly available 


## Installation

#### Pip

```bash

# [OPTIONAL] create conda environment
conda create -n pinn python=3.9
conda activate pinn

\textsc{}# install pytorch according to instructions
# https://pytorch.org/get-started/

# install requirements
pip install -r requirements.txt
```
We include the conda export 'environment.yml' file for a backup, if encounter trouble when using pip.

The installation for jaxlib can be a bit annoying and depends on your cuda version. I have used cuda 11.8. CPU-only jax is good enough for inference, so you can try that.

```
# Add jaxlib to the LD_LIBRARY_PATH, we only want to have jaxlib here -- add this line to your .bashrc or .hpc/.python.sh
export LD_LIBRARY_PATH=$VIRTUAL_ENV/lib/python3.10/site-packages/jaxlib

# Test the installation
python -c "import torch; print('PyTorch CUDA version:', torch.version.cuda); print('PyTorch cuDNN version:', torch.backends.cudnn.version())"
python -c "from jax.lib import xla_bridge; print('JAX XLA backend:', xla_bridge.get_backend().platform)"

# If any of these tests failed or GPU is unavailable, try either of these commands
pip install --upgrade torch torchvision torchaudio --index-url https://download.pytorch.org/whl/cu121
pip install --upgrade "jax[cuda12_pip]" -f https://storage.googleapis.com/jax-releases/jax_cuda_releases.html
```

## How to run
We included a checkpoint without the optimizer states for our Baseline experiment. Therefore, you should be able to run the visualization notebook now. The evaluation notebook requires you to perform step 1. 
As the pickle files might not be opening correctly for you, you can also run the data preparation notebooks beforehand. You can also train your own model.

### Prepare data and run the notebooks
1. Get the data from Zenodo: https://zenodo.org/records/11522050
2. Unzip the data and place it in the data/dorschky2024 directory
3. Prepare the training data by running the notebooks notebooks/dorschky2024_training_data_preparation.ipynb & notebooks/dorschky2024_validation_data_preparation.ipynb
   * Make sure that the paths in the notebook is the path to the data folder - they should be correct if you, for example, open the project in VSCode
   * Make sure that, after the imports. the current path is set to the root of the project
   * If not, you might need to change the paths - keep that in mind for future notebooks as well
4. The data is saved in the data/dorschky2024 directory as .pkl files
5. We have already included a checkpoint without the optimizer states for our Baseline experiment. Finetuning is possible on this, however, we have initialized the optimizer states in our finetuning experiments.
6. You can now start training as described below. The current configuration is set to the Bi-LSTM baseline experiment.
7. The evaluation notebook is also available in the notebooks directory. OMC results can be downloaded from addBiomechanics. Posting a link to the data here is not possible due to the double-blind review process. We will add that later. Therefore, the evaluation contains only JPE, speed error and jitter. 

### Training

Train model with default configuration - in the config/ folder, we have currently loaded the settings for the Bi-LSTM baseline experiment.

```bash
# train on CPU
python src/train.py trainer=cpu

# train on GPU
python src/train.py trainer=gpu
```
The built-in eval and predict scripts don't do anything in our setup - they are more suited for supervised learning tasks.

Train model with chosen experiment configuration from [configs/experiment/](configs/experiment/)

```bash
python src/train.py experiment=experiment_name.yaml
```

You can override any parameter from command line like this

```bash
python src/train.py trainer.max_epochs=20 datamodule.batch_size=64
```

## Experimentation
After training the model with the chosen experiment configuration and experiment_name the results are saved in the logs directory 
with the experiment_name. 

Several information such as the configurations, checkpoints and tensorboard files are saved under the experiment_name directory
which is created by hydra 


## Project Structure

The directory structure of new project looks like this:

```
├── configs                   <- Hydra configuration files
│   ├── callbacks                <- Callbacks configs
│   ├── datamodule               <- Datamodule configs
│   ├── debug                    <- Debugging configs
│   ├── experiment               <- Experiment configs
│   ├── extras                   <- Extra utilities configs
│   ├── hparams_search           <- Hyperparameter search configs
│   ├── hydra                    <- Hydra configs
│   ├── local                    <- Local configs
│   ├── logger                   <- Logger configs
│   ├── model                    <- Model configs
│   ├── paths                    <- Project paths configs
│   ├── trainer                  <- Trainer configs
│   │
│   ├── eval.yaml             <- Main config for evaluation
│   └── train.yaml            <- Main config for training
│
├── data                   <- Project data
│
├── logs                   <- Logs generated by hydra and lightning loggers
│
├── notebooks              <- Jupyter notebooks.
│   ├── dorschky2024_training_data_preparation.ipynb    <- Create a .pkl file for the training data
│   ├── dorschky2024_validation_data_preparation.ipynb  <- Create a .pkl file for the training data
│   └── vis_results.ipynb                               <- Notebook to create visualizations
│
├── src                    <- Source code
│   ├── datamodules              <- Lightning datamodules
│   ├── models                   <- Lightning models, "sspinn_module" contains the main SSPINNpose implementation
│   ├── utils                    <- Utility scripts
│   ├── kinematics/kinematics_2d.py <- Functions to calculate global positions
│   ├── multibody_sim            <- multibody dynamics implementation
│   │   ├── equations_of_motion.py      <- equations script
│   │   ├── ground_contact.py           <- function for ground contact modeling
│   │   ├── inertia.py                  <- inertia script
│   │   ├── kinematics.py               <- kinematics script
│   │   ├── kinetics.py                 <- kinetics script
│   │   ├── simulation.py               <- simulation script
│   │   └── utils.py                    <- utils script
│   │
│   ├── eval.py                  <- Run evaluation
│   └── train.py                 <- Run training
││
├── .env.example              <- Example of file for storing private environment variables
├── .gitignore                <- List of files ignored by git
├── .pre-commit-config.yaml   <- Configuration of pre-commit hooks for code formatting
├── requirements.txt          <- File for installing python dependencies
└── README.md
```

## Multibody Model 
The multibody dynamical model consists of the following scripts:

#### kinematics.py:
Defines the kinematic relationships (i.e. motion) among rigid bodies in the system

#### kinetics.py:
Defines several loads that act on the rigid bodies of the system

#### inertia.py:
Information about the mass and inertia of the bodies involved.

#### equations_of_motion.py:
Defines equation of motion using Kane's Method with the generalized coordinates, the generalized speeds, the kinematical differential equations, the loads, the bodies, and a Newtonian reference frame

#### simulation.py:
Defines equation of motion using Kane's Method for visualization and analyses for the notebooks scripts

#### utils.py:
Host methods for auxilary information needed for Kane's Method 

