#!/usr/bin/env python

from sympy.physics.mechanics import inertia, RigidBody

from src.multibody_sim.kinematics import *

# Mass
# ====

thigh_mass, shank_mass, foot_mass, torso_mass = symbols('m_T, m_S, m_F, m_H')

# Inertia
# =======

thigh_inertia, shank_inertia, foot_inertia, torso_inertia = symbols('I_T, I_S, I_F, I_H')

# right leg inertias
right_thigh_inertia_dyadic = inertia(right_thigh_frame, 0, 0, thigh_inertia)
right_thigh_central_inertia = (right_thigh_inertia_dyadic, right_thigh_mass_center)

right_shank_inertia_dyadic = inertia(right_shank_frame, 0, 0, shank_inertia)
right_shank_central_inertia = (right_shank_inertia_dyadic, right_shank_mass_center)

right_foot_inertia_dyadic = inertia(right_foot_frame, 0, 0, foot_inertia)
right_foot_central_inertia = (right_foot_inertia_dyadic, right_foot_mass_center)

# left leg inertias
left_thigh_inertia_dyadic = inertia(left_thigh_frame, 0, 0, thigh_inertia)
left_thigh_central_inertia = (left_thigh_inertia_dyadic, left_thigh_mass_center)

left_shank_inertia_dyadic = inertia(left_shank_frame, 0, 0, shank_inertia)
left_shank_central_inertia = (left_shank_inertia_dyadic, left_shank_mass_center)

left_foot_inertia_dyadic = inertia(left_foot_frame, 0, 0, foot_inertia)
left_foot_central_inertia = (left_foot_inertia_dyadic, left_foot_mass_center)

# torso inertias
torso_inertia_dyadic = inertia(torso_frame, 0, 0, torso_inertia)
torso_central_inertia = (torso_inertia_dyadic, torso_mass_center)



# Rigid Bodies
# ============

# right leg rigid bodies
right_thigh = RigidBody('right_thigh', right_thigh_mass_center, right_thigh_frame, thigh_mass, right_thigh_central_inertia)
right_shank = RigidBody('right_shank', right_shank_mass_center, right_shank_frame, shank_mass, right_shank_central_inertia)
right_foot = RigidBody('right_foot', right_foot_mass_center, right_foot_frame, foot_mass, right_foot_central_inertia)

# left leg rigid bodies
left_thigh = RigidBody('left_thigh', left_thigh_mass_center, left_thigh_frame, thigh_mass, left_thigh_central_inertia)
left_shank = RigidBody('left_shank', left_shank_mass_center, left_shank_frame, shank_mass, left_shank_central_inertia)
left_foot = RigidBody('left_foot', left_foot_mass_center, left_foot_frame, foot_mass, left_foot_central_inertia)

# torso rigid bodie
torso = RigidBody('torso', torso_mass_center, torso_frame, torso_mass, torso_central_inertia)