#!/usr/bin/env python

from src.multibody_sim.inertia import *

# Gravity
# =======

g = symbols('g')

# right leg gravitional forces
right_thigh_grav_force = (right_thigh_mass_center, -thigh_mass * g * ground_frame.y)
right_shank_grav_force = (right_shank_mass_center, -shank_mass * g * ground_frame.y)
right_foot_grav_force = (right_foot_mass_center, -foot_mass * g * ground_frame.y)

# left leg gravitional forces
left_thigh_grav_force = (left_thigh_mass_center, -thigh_mass * g * ground_frame.y)
left_shank_grav_force = (left_shank_mass_center, -shank_mass * g * ground_frame.y)
left_foot_grav_force = (left_foot_mass_center, -foot_mass * g * ground_frame.y)

# torso gravitational force 
torso_grav_force = (torso_mass_center, -torso_mass * g * ground_frame.y)


# Ground Reaction Force
# =======

# right ground reaction forces
grf_r_x, grf_r_y = dynamicsymbols('GRF_heel_x_r, GRF_heel_y_r')
m_r = dynamicsymbols('M_r')
right_grf = (right_ankle, grf_r_x * ground_frame.x + grf_r_y * ground_frame.y)
right_moment = (right_foot_frame, m_r * ground_frame.z)

# left ground reaction forces
grf_l_x, grf_l_y = dynamicsymbols('GRF_heel_x_l, GRF_heel_y_l')
m_l = dynamicsymbols('M_l')
left_grf = (left_ankle, grf_l_x * ground_frame.x + grf_l_y * ground_frame.y)
left_moment = (left_foot_frame, m_l * ground_frame.z)

# Old ground contact model from altan
#right_ground_force_x, right_ground_force_y = dynamicsymbols('GFx_r, GFy_r')
#right_center_of_press_x, right_center_of_press_y = dynamicsymbols('CoP_x_r, CoP_y_r')
#right_ground_reaction_point = Point('GRP_r')
#
#right_ground_reaction_point.set_pos(right_ankle, right_center_of_press_x * right_foot_frame.x
#                                    + (right_center_of_press_y) * right_foot_frame.y)
#right_ground_force = (right_ground_reaction_point, right_ground_force_x * ground_frame.x
#                                    + right_ground_force_y * ground_frame.y )
#
## left ground reaction forces
#left_ground_force_x, left_ground_force_y = dynamicsymbols('GFx_l, GFy_l')
#left_center_of_press_x, left_center_of_press_y = dynamicsymbols('CoP_x_l, CoP_y_l')
#left_ground_reaction_point = Point('GRP_l')
#
#left_ground_reaction_point.set_pos(left_ankle, left_center_of_press_x * left_foot_frame.x
#                                    + (left_center_of_press_y) * left_foot_frame.y)
#left_ground_force = (left_ground_reaction_point, left_ground_force_x * ground_frame.x  # maybe define x & y forces seperatly
#                                    + left_ground_force_y * ground_frame.y)



# Joint Torques
# =============

# right leg torques 
right_hip_torque, right_knee_torque, right_ankle_torque = dynamicsymbols('T_h_r, T_k_r, T_a_r')
right_thigh_torque = (right_thigh_frame,  right_hip_torque * ground_frame.z)
right_shank_torque = (right_shank_frame, right_knee_torque * ground_frame.z)
right_foot_torque = (right_foot_frame,  right_ankle_torque * ground_frame.z)

# left leg torques 
left_hip_torque, left_knee_torque, left_ankle_torque = dynamicsymbols('T_h_l, T_k_l, T_a_l')
left_thigh_torque = (left_thigh_frame, left_hip_torque * ground_frame.z)
left_shank_torque = (left_shank_frame, left_knee_torque * ground_frame.z)
left_foot_torque = (left_foot_frame, left_ankle_torque * ground_frame.z)


# # right leg torques 
# right_hip_torque, right_knee_torque, right_ankle_torque = dynamicsymbols('T_h_r, T_k_r, T_a_r')
# right_thigh_torque = (right_thigh_frame, +right_knee_torque * ground_frame.z + right_hip_torque * ground_frame.z) 
# right_shank_torque = (right_shank_frame, -right_ankle_torque * ground_frame.z + -right_knee_torque * ground_frame.z)
# right_foot_torque = (right_foot_frame, +right_ankle_torque * ground_frame.z)

# # left leg torques 
# left_hip_torque, left_knee_torque, left_ankle_torque = dynamicsymbols('T_h_l, T_k_l, T_a_l')
# left_thigh_torque = (left_thigh_frame, +left_knee_torque * ground_frame.z + left_hip_torque * ground_frame.z) 
# left_shank_torque = (left_shank_frame, -left_ankle_torque * ground_frame.z + -left_knee_torque * ground_frame.z)
# left_foot_torque = (left_foot_frame, +left_ankle_torque * ground_frame.z)

# # torso torque
# torso_torque = (torso_frame, -right_hip_torque * ground_frame.z -left_hip_torque * ground_frame.z)



