from typing import Iterable, Union
import hydra
from omegaconf import OmegaConf
import omegaconf


# Custom resolver to calculate the length of a list
def full_rec_length(item) -> int:
    if isinstance(item, (list, tuple, set, omegaconf.listconfig.ListConfig)):
        return sum(full_rec_length(v) for v in item)
    if isinstance(item, (dict, omegaconf.dictconfig.DictConfig)):
        return sum(full_rec_length(v) for k, v in item.items())
    return 1


def register_resolvers(log):
    if log is not None:
        log.info("Registering custom input_dict resolver")
    OmegaConf.register_new_resolver("full_rec_length", full_rec_length)
